"""
Reinforce implementation in TensorFlow 2. Main function is main.py
"""
import numpy as np
import os
import pickle
import tensorflow as tf
import matplotlib.pyplot as plt
from policy  import  Policy
import ipdb
import copy
import random
from keras import backend as K
from keras.models import model_from_json
from Environments.PathoMC import PathoMcar as env_car
import tensorflow_probability as tfp



def export_plot(ys, ylabel, title, filename):
    plt.figure()
    plt.plot(range(len(ys)), ys)
    plt.xlabel("Training Episode")
    plt.ylabel(ylabel)
    plt.title(title)
    plt.savefig(filename)
    plt.close()

"""
the following class implements policy gradient
"""

class PolicyGradient(object):
    def __init__(self, env, action_space_type, max_ep_len=200,  num_iterations=1000, batch_size=2,  output_path="../Dummy_env/results/"):

        self.output_path = output_path
        if not os.path.exists(output_path):
            os.makedirs(output_path)
        self.env = env
        self.step_size = env["step_size"]

        self.action_space_type = action_space_type
        self.observation_dim = env["observation_dim"] #self.env.observation_space.shape[0]

        if (self.action_space_type == "continuous" or self.action_space_type == "continuous_const_variance"): 
            self.action_dim = env["action_dim"] # self.env.action_space.shape[0]
        self.gamma = 0.99
        self.num_iterations = num_iterations
        self.policy_type = env["distribution"]
        self.action_limit = env["action_limit"]
        self.batch_size = batch_size
        self.max_ep_len = max_ep_len
        self.lr_schedule = tf.keras.optimizers.schedules.ExponentialDecay(
            initial_learning_rate=env["step_size"],
            decay_steps=5000,
            decay_rate=0.9)
        self.optimizer = tf.keras.optimizers.SGD(learning_rate=self.lr_schedule, clipvalue=10.0)
        self.policy = Policy(input_size=self.observation_dim, output_size=self.action_dim, action_space_type=self.action_space_type, env=self.env)

    """
    this function is to train and update the parameters
    """
    def train(self):

        all_total_rewards = []
        all_total_rewards_prev =[]

        averaged_total_rewards = []
        averaged_total_rewards_cum = []
        iterations = []
        gradient_stored =[]
        gradient_stored_cum =[]

        for t in range(self.num_iterations):
            iterations.append(t)
            if t == 0:
                paths, total_rewards, paths_prev,total_rewards_prev,loaded_model_prev = self.play_games(epoch=t)
            
            else: #flattening the list
                paths, total_rewards,paths_prev,total_rewards_prev,loaded_model_prev = self.play_games(epoch=t)  ##numpy
                all_total_rewards_prev.extend(total_rewards_prev) #numpy
                observations_prev = np.concatenate([path_prev["observation_prev"] for path_prev in paths_prev]) #numpy
                actions_prev = np.concatenate([path_prev["action_prev"] for path_prev in paths_prev]) #numpy
                returns_prev = self.get_returns(paths_prev,0) #numpy

            all_total_rewards.extend(total_rewards)
            observations = np.concatenate([path["observation"] for path in paths])
            actions = np.concatenate([path["action"] for path in paths])
            returns = self.get_returns(paths,1)
            """
            The following if condition takes care of initial g_k_1 (g_0 in Algorithm 1)
            """
            if t == 0:
                ########################## need to check here ########################################
                observations_prev, actions_prev, returns_prev = [], [], []
                g_k_1,g_k_cap = self.update_policy(observations, actions, returns, observations_prev, actions_prev, returns_prev,loaded_model_prev,t,[])
                g_k_caps = g_k_cap
            else:
                g_k_1,g_k_cap = self.update_policy(observations, actions, returns, observations_prev, actions_prev, returns_prev, loaded_model_prev, t, g_k_1)
                g_k_caps = g_k_cap
            avg_reward = np.mean(total_rewards)
            averaged_total_rewards.append(avg_reward)
            gradient_stored.append(g_k_caps)
            gradient_stored_cum.append(np.mean(gradient_stored[-100:]))

            averaged_total_rewards_cum.append(np.mean(averaged_total_rewards[-100:])) # running avg
            print("Average reward for batch {}: {:04.2f}".format(t,gradient_stored_cum[-1]), averaged_total_rewards_cum[-1])
        """
        Options to plot and save results
        """
        print("Training complete")
        """
        Total reward
        """
        np.save(self.output_path+ "rewards.npy", averaged_total_rewards)
        export_plot(averaged_total_rewards, "Reward", "Continuous_Car", self.output_path + "reward.png")
        """
        Cumulative reward
        """
        np.save(self.output_path + "cum_rewards.npy", averaged_total_rewards_cum)
        export_plot(averaged_total_rewards_cum, "Reward", "Continuous_Car", self.output_path + "Reward_running_avg.png")
        """
        Gradient averaged
        """
        np.save(self.output_path + "Gradient_cum.npy", gradient_stored)
        export_plot(gradient_stored_cum, "Averaged Gradient", "Continuous_Car", self.output_path + "average_gradients.png")
        pickle.dump([averaged_total_rewards_cum, averaged_total_rewards, gradient_stored_cum, gradient_stored, iterations],
                    open(self.output_path + "results.p", "wb"))

    """
    Generate roll-outs for theta_k and theta_k-1
    """

    def play_games(self, epoch, env=None, num_episodes = None):
        episode = 0
        episode_rewards = []
        paths = []
        # t = 0

        episode_prev = 0
        episode_rewards_prev = []
        paths_prev = []
        # t_prev = 0

        while episode < self.batch_size:
            """
            Roll-out old trajectory for theta_k
            """

            state = np.array([2.5]) #([random.uniform(1.15,2.0)]) #np.array([0.0]) #env.reset()

            states, actions, rewards = [], [], []
            episode_reward = 0

            for step in range(self.max_ep_len):
                states.append(state)
                action = self.policy.sampl_action(state)[0]  #### sampl_action returns numpy
                state, reward, done = env_car(state,action)  #### returns numpy, all state actions are numpy objects
                #env.step(action)

                actions.append(action)
                rewards.append(reward)
                episode_reward += reward
                # t += 1

                if (done or step == self.max_ep_len-1):

                    episode_rewards.append(episode_reward)
                    break

            path = {"observation": np.array(states),
                    "reward": np.array(rewards),
                    "action": np.array(actions)}
            paths.append(path)

            #load junk model
            json_file = open('/nfshomes/schakra3/Algorthm_1/Dummy_env/junk/model_prev.json', 'r')
            loaded_model_json = json_file.read()
            json_file.close()
            loaded_model = model_from_json(loaded_model_json)
            loaded_model.load_weights('/nfshomes/schakra3/Algorthm_1/Dummy_env/junk/model_prev.h5')

            """
            Roll-out old trajectory for theta_k-1 for all episodes> 0
            """
            if epoch != 0:
                """
                Load previous model
                """
                json_file = open(self.output_path + 'model_prev.json', 'r')
                loaded_model_json = json_file.read()
                json_file.close()
                loaded_model = model_from_json(loaded_model_json)
                
                # load weights into new model
                loaded_model.load_weights(self.output_path + "model_prev.h5")

                state_prev =  np.array([2.5]) #np.array([random.uniform(1.15,2.0)])  # env.reset()
                states_prev, actions_prev, rewards_prev = [], [], []
                episode_reward_prev = 0

                for step_prev in range(self.max_ep_len):
                    states_prev.append(state_prev)
                    action_prev = self.sampl_action(np.atleast_2d(state_prev),loaded_model)[0]
                    state_prev, reward_prev, done_prev = env_car(state_prev, action_prev)
                    # env.step(action)

                    actions_prev.append(action_prev)
                    rewards_prev.append(reward_prev)
                    episode_reward_prev += reward_prev
                    # t_prev += 1

                    if done_prev or step_prev == self.max_ep_len - 1:
                        episode_rewards_prev.append(episode_reward_prev)
                        break

                path_prev = {"observation_prev": np.array(states_prev),
                             "reward_prev": np.array(rewards_prev),
                             "action_prev": np.array(actions_prev)}
                paths_prev.append(path_prev)

            episode += 1

            if episode > self.batch_size-1: #num_episodes:
                break

        #returns all operations in numpy
        return paths, episode_rewards, paths_prev, episode_rewards_prev, loaded_model
        

    """
    To sample observations using theta_k-1
    """
    def action_distribution_prev(self, observations_prev,loaded_model):
        
        if self.action_space_type =="discrete":
            logits = loaded_model(observations_prev)
            return tfp.distributions.Categorical(logits=logits)

        elif self.action_space_type == "continuous":
            
            if self.policy_type =="Cauchy":
                [loc, scale] = loaded_model(observations_prev)
                """
                For actions in multi-dimension, use tfp.distributions.MultivariateNormalDiag()
                """
                return tfp.distributions.Cauchy(loc, scale)
            else:
                [loc, scale] = loaded_model(observations_prev)
                """
                For actions in multi-dimension, use tfp.distributions.MultivariateNormalDiag()
                """
                return tfp.distributions.Normal(loc, scale)

        elif self.action_space_type == "continuous_const_variance":
        
            if self.policy_type =="Cauchy":
                loc  = loaded_model(observations_prev)
                """
                For actions in multi-dimension, use tfp.distributions.MultivariateNormalDiag()
                """
                return tfp.distributions.Cauchy(loc, scale = 0.5)
            
            else:
                loc = loaded_model(observations_prev)
                """
                For actions in multi-dimension, use tfp.distributions.MultivariateNormalDiag()
                """
                return tfp.distributions.Normal(loc, scale= 0.5)



    def sampl_action(self, observations,loaded_model):
        sampled_actions = self.action_distribution_prev(observations,loaded_model).sample()
        
        if self.action_space_type == "continuous":
            action = K.clip(sampled_actions, -self.action_limit, self.action_limit)
            return action.numpy()

        elif self.action_space_type == "continuous_const_variance":
            #action is clipped within a value
            action = K.clip(sampled_actions, -self.action_limit, self.action_limit)
            return action.numpy()

        return sampled_actions.numpy()   ##### returns in numpy

    """
    Find expected returns for given paths
    """
    
    def get_returns(self, paths, token):
        all_returns = []
        for path in paths:
            if token == 1:
                rewards = path["reward"]
            if token == 0:
                rewards = path["reward_prev"]
            returns = []
            reversed_rewards = np.flip(rewards, 0)
            g_t = 0
            for r in reversed_rewards:
                g_t = r + self.gamma*g_t
                returns.insert(0, g_t)
            all_returns.append(returns)
        returns = np.concatenate(all_returns)
        returns = (returns - np.mean(returns)) / (np.std(returns) + 1e-7)
        return returns   #### returns in numpy

    def update_policy(self, observations, actions, returns,observations_prev,actions_prev,returns_prev,loaded_model_prev,t,g_k_1=[]):
        """
        This part updates the parameters
        :param observations: states corresponds to theta_k (xi_k(theta_k)) of (5)
        :param actions: corresponds to theta_k
        :param returns: corresponds to theta_k
        :param observations_prev: states corresponds to theta_k-1 (xi_k(theta_k-1)) of (5)
        :param actions_prev: corresponds to theta_k-1
        :param returns_prev: corresponds to theta_k-1
        :param g_k_1: g_{k-1} of step 4, Algorithm 1
        :return:  updated g_k_1 for step 4 and also updates theta_k
        """
        step_size = (3e-3) * (0.9) ** (t / 5000)

        #check : used deep copy
        #theta_0 = copy.copy(self.policy.model.get_weights()) # [i[0].numpy() for i in zip(self.policy.model.trainable_weights)]
        #theta_0 : old model
        theta_0 = copy.deepcopy(self.policy.model.get_weights())

        if observations_prev == [] and actions_prev == [] and returns_prev == []:
            observations = tf.convert_to_tensor(observations) #tensor conversion
            actions = tf.convert_to_tensor(actions)
            returns = tf.convert_to_tensor(returns)
            for obs, act, ret in zip(observations,actions,returns):
                
                with tf.GradientTape() as tape: ### so that we can take gradiuents
                    log_prob = self.policy.action_distribution(obs).log_prob(act) + 1e-5  ### tf distribution object
                    loss = -tf.math.reduce_mean(log_prob * tf.cast(ret, tf.float32))     ## tf tensor

                deltaj_k = tape.gradient(loss, self.policy.model.trainable_weights)    ### tf
                
                
                with tf.GradientTape() as tape:

                    #chec
                    g_k0 = tf.convert_to_tensor(
                        sum((i * j) for i, j in zip(deltaj_k, self.policy.model.trainable_weights))\
                        + 0.000001 * (1/step_size) * sum([(i-j)*(i-j) for i,j in zip(self.policy.model.trainable_weights, theta_0)]))
                    

                gradient_final = tape.gradient(g_k0, self.policy.model.trainable_weights)
                
                self.optimizer.apply_gradients(zip(gradient_final, self.policy.model.trainable_weights))
            
            
            g_k_1 = copy.deepcopy(deltaj_k)
            
            """
            Save model weights for next step (k+1) 
            """
            model_json_prev = self.policy.model.to_json()
            
            with open(self.output_path + "model_prev.json", "w") as json_file:
                json_file.write(model_json_prev)
            self.policy.model.save_weights(self.output_path + "model_prev.h5")
            del tape

            #tf operations
            g_k_caps = (1/step_size)* np.linalg.norm(np.array(self.policy.model.get_weights())-np.array(theta_0))

            return g_k_1, g_k_caps

        else:
            observations_prev = tf.convert_to_tensor(observations_prev)
            actions_prev = tf.convert_to_tensor(actions_prev)
            returns_prev = tf.convert_to_tensor(returns_prev)
            observations = tf.convert_to_tensor(observations)
            actions = tf.convert_to_tensor(actions)
            returns = tf.convert_to_tensor(returns)

            for obs, act, ret, obs_prev, act_prev, ret_prev in zip(observations, actions, returns, observations_prev, actions_prev, returns_prev):
                
                with tf.GradientTape(persistent=True) as tape:
                    log_prob = self.policy.action_distribution(obs).log_prob(act) + 1e-5
                    loss = -tf.math.reduce_mean(log_prob * tf.cast(ret, tf.float32))

                    #need to correct here
                    # log_prob_prev = self.policy.action_distribution(obs_prev).log_prob(act_prev) + 1e-5
                    log_prob_prev = self.action_distribution_prev(obs_prev, loaded_model_prev).log_prob(act_prev) + 1e-5
                    loss_prev = -tf.math.reduce_mean(log_prob_prev * tf.cast(ret_prev, tf.float32))


                deltaj_k = tape.gradient(loss, self.policy.model.trainable_weights)
                # deltaj_k_1 = tape.gradient(loss_prev, self.policy.model.trainable_weights)
                deltaj_k_1 = tape.gradient(loss_prev, loaded_model_prev.trainable_weights)

                #

                del tape
                
                with tf.GradientTape(persistent=True) as tape:
                    #modify
                    g_k = [0.0 * (z - j) + i for i, j, z in zip(deltaj_k, deltaj_k_1, g_k_1)]

                    #need to comp
                    
                    ### tf.math.reduce_sum(tf.tensordot(a, b))
                    g_k0 = sum(i * j for i, j in zip(g_k, self.policy.model.trainable_weights))\
                    + 0.000001 * (1/step_size) * sum([(i-j)*(i-j) for i,j in zip(self.policy.model.trainable_weights, theta_0)])

                gradient_final = tape.gradient(g_k0, self.policy.model.trainable_weights)
                
                #implements gradient descent NOT ASCENT : Fix Sign
                # print (gradient_final)
                self.optimizer.apply_gradients(zip(gradient_final, self.policy.model.trainable_weights))
            
            g_k_1 = copy.deepcopy(g_k)
            del tape

            #use tensorflow tf instead of np
            g_k_caps = (1/step_size)*np.linalg.norm(np.array(self.policy.model.get_weights())-np.array(theta_0))
            return g_k_1, g_k_caps



   