import tensorflow_probability as tfp
import tensorflow as tf
from tensorflow import keras
from keras import backend as K
from tensorflow.keras import layers
from tensorflow.keras.layers import  Dense, Input
from tensorflow.keras.models import Model
import sklearn.preprocessing
import numpy as np
import ipdb


class Policy():
    def __init__(self, input_size, output_size, action_space_type, env):
        self.action_space_type = action_space_type
        self.action_limit = env["action_limit"]
        self.distribution = env["distribution"]
    

        if self.action_space_type == "discrete":
            self.model = keras.Sequential(
                layers=[
                    keras.Input(shape=(input_size,)),
                    layers.Dense(64, activation="relu", name="relu_layer"),
                    layers.Dense(output_size, activation="linear", name="linear_layer")
                ],
                name="policy")

        elif self.action_space_type == "continuous":
            inputs = Input(shape=(input_size,), name='state')
            layer1 = Dense(1, activation="tanh",kernel_initializer=tf.keras.initializers.random_normal(seed=0),
                           name="elu_layer1")(inputs)
            layer2 = Dense(1, activation="elu",kernel_initializer=tf.keras.initializers.random_normal(seed=0),
                           name="elu_layer2")(layer1)
            mean = Dense(output_size, activation='linear',
                         name='mean')(layer2)

            stddev = Dense(output_size,activation='linear',
                           name='stddev')(inputs)
            stddev = K.softplus(stddev) + 1e-5
            # use of softplusk avoids stddev = 0

            self.model = Model(inputs, [mean, stddev], name='policy')
            self.model.summary()
        

       
        elif self.action_space_type == "continuous_const_variance":
            inputs = Input(shape=(input_size,), name='state')
            layer1 = Dense(1, activation="tanh",kernel_initializer=tf.keras.initializers.random_normal(seed=0),
                           name="elu_layer1")(inputs)
            layer2 = Dense(1, activation="elu",kernel_initializer=tf.keras.initializers.random_normal(seed=0),
                           name="elu_layer2")(layer1)
            mean = Dense(output_size, activation='linear',
                         name='mean')(layer2)

            self.model = Model(inputs, mean, name='policy')
            self.model.summary()



    def action_distribution(self, observations):
        
        if self.action_space_type == "discrete":
            logits = self.model(observations)
            return tfp.distributions.Categorical(logits=logits)
        
        elif self.action_space_type == "continuous":
            
            if self.distribution == "Cauchy":
                [loc, scale] = self.model(observations)
                """
                For actions in multi-dimension, use tfp.distributions.MultivariateNormalDiag()
                """

                return tfp.distributions.Cauchy(loc, scale)

            else:
                [loc, scale] = self.model(observations)
                """
                For actions in multi-dimension, use tfp.distributions.MultivariateNormalDiag()
                """

                return tfp.distributions.Normal(loc, scale)

        
        #constant variance case
        elif self.action_space_type == "continuous_const_variance":
            
            if self.distribution == "Cauchy":
                loc = self.model(observations)
                """
                For actions in multi-dimension, use tfp.distributions.MultivariateNormalDiag()
                """

                return tfp.distributions.Cauchy(loc, scale=0.1)

            if self.distribution == "Gaussian":
                loc = self.model(observations)
                
                """
                For actions in multi-dimension, use tfp.distributions.MultivariateNormalDiag()
                """

                return tfp.distributions.Normal(loc, scale=0.1)



    def sampl_action(self, observations):
        #get sampled act
        sampled_actions = self.action_distribution(observations).sample()
        
        if self.action_space_type == "continuous":
            action = K.clip(sampled_actions, -self.action_limit, self.action_limit) #
            return action.numpy()

        elif self.action_space_type == "continuous_const_variance":
            action = K.clip(sampled_actions, -self.action_limit, self.action_limit) #
            return action.numpy()

        return sampled_actions.numpy()






